from setuptools import setup

setup(name='cb4_reports_exporter',
      version='1.0.4',
      description='Export reports from CB4',
      author='CB4',
      author_email="support@cb4.com",
      packages=['reports_exporter'],
      python_requires='>=3.7',
      zip_safe=False)
